package cz.mendelu.dnaAnalyser.analyse.cpg;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;


@Slf4j
@Service
public class CpgAnalyser {

    private static double MIN_GC_PERCENTAGE;
    private static double MIN_OBSERVED_TO_EXPECTED_CPG;
    private static int MIN_WINDOW_SIZE;
    private static int MIN_ISLAND_MERGE_GAP;
    private static char FIRST_NUCLEOTIDE;
    private static char SECOND_NUCLEOTIDE;

    private static boolean findIsland1BpShifts(CpgWindow window) {
        while(true) {
            if (window.isIsland()) return true;
            if (!window.expandRight()) return false;
            window.shrinkLeft();
        }
    }

    private static boolean twoFoldShrinking(CpgWindow window) {
        while(!window.isIsland()) {
            window.shrinkRight();
            window.shrinkLeft();
            window.windowSize = window.windowSize - 2;
            if (window.windowSize < MIN_WINDOW_SIZE) return false;
        }
        return true;
    }

    private static ArrayList<CpgWindow> mergeIslands(ArrayList<CpgWindow> foundIslands) {
        ArrayList<CpgWindow> mergedIslands = new ArrayList<>();
        mergedIslands.add(foundIslands.get(0));
        if (foundIslands.size() >= 2) {
            for (CpgWindow foundIsland : foundIslands.subList(1, foundIslands.size())) {
                if (foundIsland.windowBegin - mergedIslands.get(mergedIslands.size() - 1).windowEnd < MIN_ISLAND_MERGE_GAP) {
                    mergedIslands.get(mergedIslands.size() - 1).joinWindows(foundIsland);
                } else {
                    mergedIslands.add(foundIsland);
                }
            }
        }
        return mergedIslands;
    }

    private static void extendIslandWindowShifts(CpgWindow window) {
        while(true) {
            boolean extended = window.expandRightIsland();
            if (!extended) break;
        }
        window.updateGcCount();
    }

    private static ArrayList<CpgWindow> findIslands(String sequence) {
        if (sequence.length() < MIN_WINDOW_SIZE) {
            return new ArrayList<>();
        }
        ArrayList<CpgWindow> foundIslands = new ArrayList<>();
        int recordPosition = 0;
        while(true) {
            CpgWindow window = new CpgWindow(sequence, recordPosition, MIN_WINDOW_SIZE, MIN_GC_PERCENTAGE, MIN_OBSERVED_TO_EXPECTED_CPG, MIN_WINDOW_SIZE, FIRST_NUCLEOTIDE, SECOND_NUCLEOTIDE);
            // step 1 find window that is an island
            boolean recordContinues = findIsland1BpShifts(window);
            if (!recordContinues) break;

            // step 2 window-length shifts
            extendIslandWindowShifts(window);

            // step 3 shift the last window by 1bp until it meets the criteria
            window.rollbackUntilIsland();

            // step 4 shrink the whole island by 1bp until it meets the criteria
            if(!twoFoldShrinking(window)){
                recordPosition = window.windowBegin + 1;
                continue;
            }
            foundIslands.add(window);
            recordPosition = window.windowEnd;
            if (recordPosition + MIN_WINDOW_SIZE >= sequence.length()) break;
        }
        // step 5 at the end merge islands that are at least 100bp apart
        if (foundIslands.isEmpty()) return foundIslands;
        return mergeIslands(foundIslands);
    }

    private void initialize(int minWindowSize, double minGcPercentage, double minObservedToExpectedCpG, int minIslandMergeGap, char firstNucleotide, char secondNucleotide) {
        MIN_GC_PERCENTAGE = minGcPercentage;
        MIN_OBSERVED_TO_EXPECTED_CPG = minObservedToExpectedCpG;
        MIN_WINDOW_SIZE = minWindowSize;
        MIN_ISLAND_MERGE_GAP = minIslandMergeGap;
        FIRST_NUCLEOTIDE = firstNucleotide;
        SECOND_NUCLEOTIDE = secondNucleotide;
    }

    public ArrayList<CpgAnalyseResult> getResults(String sequence, int minWindowSize, double minGcPercentage, double minObservedToExpectedCpG, int minIslandMergeGap, char firstNucleotide, char secondNucleotide) {
        initialize(minWindowSize, minGcPercentage, minObservedToExpectedCpG, minIslandMergeGap, firstNucleotide, secondNucleotide);
        ArrayList<CpgWindow> islands = findIslands(sequence);
        ArrayList<CpgAnalyseResult> results = new ArrayList<>();
        for (CpgWindow island : islands) {
            results.add(new CpgAnalyseResult(null, island.windowBegin, island.windowEnd, island.record.substring(island.windowBegin, island.windowEnd), island.getGcPerc(), island.getObsExp(), island.windowSize));
        }
        return results;
    }
}
