package cz.mendelu.dnaAnalyser.analyse.rloopr;

import cz.mendelu.dnaAnalyser.analyse.g4hunter.G4Hunter;
import cz.mendelu.dnaAnalyser.analyse.g4hunter.quadruplex.AggregateQuadruplex;
import cz.mendelu.dnaAnalyser.analyse.g4hunter.quadruplex.AggregateQuadruplexMapper;
import cz.mendelu.dnaAnalyser.batch.BatchService;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceData;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataRepository;
import cz.mendelu.dnaAnalyser.sequence.stream.BufferedWindow;
import cz.mendelu.dnaAnalyser.sequence.stream.Window;
import cz.mendelu.dnaAnalyser.utils.analyse.model.Analyse;
import cz.mendelu.dnaAnalyser.utils.analyse.service.heatmap.Heatmap;
import cz.mendelu.dnaAnalyser.utils.exception.NotFoundException;
import cz.mendelu.dnaAnalyser.utils.repository.BaseRepository;
import cz.mendelu.dnaAnalyser.utils.service.ModelService;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.session.RowBounds;
import org.apache.ibatis.session.SqlSession;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Sort;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.io.IOException;
import java.sql.*;

@Service
@Slf4j
public class RlooprService extends ModelService<Rloop> {

    @Autowired
    private RlooprRepository rlooprRepository;

    @Autowired
    private SequenceDataRepository sequenceDataRepository;

    @Autowired
    private BatchService batchService;

    @Autowired
    private RloopR rloopr;

    @Autowired
    private RloopDataSessionService rloopDataSessionService;

    @Override
    protected BaseRepository<Rloop, UUID> getRepository() {
        return rlooprRepository;
    }

    public Rloop executeRloopAnalysis(Rloop rloop) {

        rloop = rlooprRepository.save(rloop);
        batchService.createBatch(rloop.getOwner(), rloop)
                .execute(uuid -> {
                    Rloop analyse = rlooprRepository
                            .findById(uuid)
                            .orElseThrow(() -> new NotFoundException(Rloop.class, uuid));
                    SequenceData sequenceData = sequenceDataRepository.load(analyse.getSequence());
                    String sequenceDataPlain = sequenceData.toPlain();
                    Window sequenceWindow = BufferedWindow.wrap(sequenceDataPlain);
                    Short model = analyse.getModel();

                    try (SqlSession sqlSession = rloopDataSessionService.openDataSession(analyse)) {
                        RloopMapper rloopMapper = sqlSession.getMapper(RloopMapper.class);
                        List<RlooprResult> result;

                        // FIXME: calling it multiple times is just going to prolong the computation, find a better approach
                        ArrayList<RlooprResult> result_model1 = new ArrayList<RlooprResult>();
                        ArrayList<RlooprResult> result_model2 = new ArrayList<RlooprResult>();
                        ArrayList<RlooprResult> result_model1_complement = new ArrayList<RlooprResult>();
                        ArrayList<RlooprResult> result_model2_complement = new ArrayList<RlooprResult>();

                        if (model==0) {
                            result_model1 = rloopr.getInitiationZones(model, sequenceWindow, false);
                            result_model1_complement = rloopr.getInitiationZones(model, sequenceWindow, true);
                        }
                        else if (model==1) {
                            result_model2 = rloopr.getInitiationZones(model, sequenceWindow, false);
                            result_model2_complement = rloopr.getInitiationZones(model, sequenceWindow, true);
                        }
                        else {
                            result_model1 = rloopr.getInitiationZones((short)0, sequenceWindow, false);
                            result_model1_complement = rloopr.getInitiationZones((short)0, sequenceWindow, true);
                            result_model2 = rloopr.getInitiationZones((short)1, sequenceWindow, false);
                            result_model2_complement = rloopr.getInitiationZones((short)1, sequenceWindow, true);
                        }

                        // combine results alltogether
                        result = Stream.of(result_model1, result_model1_complement, result_model2, result_model2_complement)
                                            .flatMap(x -> x.stream())
                                            .collect(Collectors.toList());
                        
                        result.forEach((single_rloop) -> { 
                            rloopMapper.insert(single_rloop);
                        });

                        analyse.setResultCount(Long.valueOf(result.size()));
                        analyse.finishNow();
                        sqlSession.commit();
                    }
                    rlooprRepository.save(analyse);
                }).start();
        return rloop;
    }

    public Page<RlooprResult> returnRloopAnalysis(UUID id, Integer sequenceStart, Integer sequenceLength, Pageable pageable) {
        Rloop analyse = findOne(id);

        int start = (sequenceStart == null) ? 0 : sequenceStart;
        int end = start + ((sequenceLength == null) ? analyse.getSequence().getLength() : sequenceLength);

        try (SqlSession sqlSession = rloopDataSessionService.openDataSession(analyse)) {
            RloopMapper rloopMapper = sqlSession.getMapper(RloopMapper.class);

            RowBounds rowBounds = new RowBounds((int) pageable.getOffset(), pageable.getPageSize());
            String orderName = "POSITION";
            Sort.Direction orderDirection = Sort.Direction.ASC;
            Sort.Order order = getSortOrderOrNull(pageable);
            if (order != null) {
                orderName = order.getProperty().toUpperCase();
                orderDirection = order.getDirection();
            }

            List<RlooprResult> list = (orderDirection == Sort.Direction.ASC)
                    ? rloopMapper.getAllBetweenPositionOrderAsc(start, end, orderName, rowBounds)
                    : rloopMapper.getAllBetweenPositionOrderDesc(start, end, orderName, rowBounds);
            int total = rloopMapper.countAllBetweenPosition(start, end);

            Page<RlooprResult> page = new PageImpl<>(list, pageable, total);            
            return page;
        }
    }

    public List<RlooprResult> getRloopAnalysis(Rloop analyse){
        List<RlooprResult> list;
        try (SqlSession sqlSession = rloopDataSessionService.openDataSession(analyse)) {
            RloopMapper rloopMapper = sqlSession.getMapper(RloopMapper.class);
            list = rloopMapper.getAll();
        }
        return list;
    }

    public Heatmap getHeatmap(UUID id, Integer segmentsCount, Integer from, Integer to) {
        Rloop analyse = findOne(id);
        Heatmap heatmap = rloopDataSessionService.heatmap(RloopMapper.class, analyse, segmentsCount, from, to);
        return heatmap;
    }

    public Rloop modifyTags(UUID id, Set<String> tags) {
        Rloop analyse = rlooprRepository.findById(id)
                .orElseThrow(() -> new NotFoundException(Rloop.class, id));
        analyse.setTags(tags);
        rlooprRepository.save(analyse);
        return analyse;
    }

    private Sort.Order getSortOrderOrNull(Pageable pageable) {
        if (pageable == null) return null;
        if (pageable.getSort() == null) return null;
        if (pageable.getSort().iterator() == null) return null;
        Iterator<Sort.Order> orderIterator = pageable.getSort().iterator();
        if (orderIterator.hasNext() == false) return null;
        return orderIterator.next();
    }


    public Integer avgRloopLen(UUID id) {
        Rloop analyse = findOne(id);
        Integer result = 0;
        
        try (SqlSession sqlSession = rloopDataSessionService.openDataSession(analyse)) {
            RloopMapper rloopMapper = sqlSession.getMapper(RloopMapper.class);
            ArrayList<RlooprResult> anaysisRloops = rloopMapper.getAll();
            Integer totalRloopLen = 0;

            for (RlooprResult rloop : anaysisRloops) {
                totalRloopLen += rloop.riz.length() + rloop.linker.length() + rloop.rez.length();
            }

           
            try {
                result = totalRloopLen / anaysisRloops.size();
            } catch (java.lang.ArithmeticException e) {
                return 0; // no rloops in analysis
            }
        }

        return result;
    }

    public Integer totalRloopLen(UUID id, String model) {
        Rloop analyse = findOne(id);
        Integer totalRloopLen = 0;
        
        try (SqlSession sqlSession = rloopDataSessionService.openDataSession(analyse)) {
            RloopMapper rloopMapper = sqlSession.getMapper(RloopMapper.class);
            ArrayList<RlooprResult> anaysisRloops = rloopMapper.getAll();

            for (RlooprResult rloop : anaysisRloops) {
                if (rloop.model.contains(model)) {
                    totalRloopLen += rloop.riz.length() + rloop.linker.length() + rloop.rez.length();
                }
            }

        }

        return totalRloopLen;
    }

}
