package cz.mendelu.dnaAnalyser.analyse.zdna;

import cz.mendelu.dnaAnalyser.jwt.JwtAuthenticationUtils;
import cz.mendelu.dnaAnalyser.jwt.JwtTokenService;
import cz.mendelu.dnaAnalyser.sequence.SequenceService;
import cz.mendelu.dnaAnalyser.exporter.ExporterService;
import cz.mendelu.dnaAnalyser.utils.analyse.service.heatmap.Heatmap;
import cz.mendelu.dnaAnalyser.utils.controller.DeleteMapping;
import cz.mendelu.dnaAnalyser.utils.controller.GetMapping;
import cz.mendelu.dnaAnalyser.utils.controller.*;
import cz.mendelu.dnaAnalyser.utils.controller.PostMapping;
import cz.mendelu.dnaAnalyser.utils.controller.PutMapping;
import cz.mendelu.dnaAnalyser.utils.swagger.ApiPaginationParams;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.Data;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.ContentDisposition;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.Authentication;
import org.springframework.web.bind.annotation.*;
import springfox.documentation.annotations.ApiIgnore;
import cz.mendelu.dnaAnalyser.utils.analyse.data.ExporterEnums.Analysis;
import cz.mendelu.dnaAnalyser.utils.analyse.data.ExporterEnums.Output;

import javax.validation.Valid;
import javax.validation.constraints.NotNull;

import java.util.Set;
import java.util.UUID;

import lombok.extern.slf4j.Slf4j;

@Slf4j
@RestController
@Api(tags = "api/analyse/zdna")
@RequestMapping("api/analyse/zdna")
public class ZdnaController {

    @Data
    static class ZdnaRequest {
        private UUID sequence;

        private Set<String> tags;

        @NotNull
        private Integer minSequenceSize = 10;

        @NotNull
        private float threshold = 0;

        @NotNull
        private float score_gc = 1;

        @NotNull
        private float score_gtac = 0.5f;

        @NotNull
        private float score_at = 0.25f;

        @NotNull
        private float score_oth = 0;

    }

    @Data
    static class HeatMapRequest {
        private Integer sequenceLength;
        private UUID sequenceId;
    }

    @Autowired
    private ExporterService exporterService;

    @Autowired
    private SequenceService sequenceService;

    @Autowired
    private JwtTokenService jwtTokenService;

    @Autowired
    private ZdnaService zdnaService;

    @PostMapping()
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Analyse sequence for Z-DNAs")
    public Response<Zdna> analyzeZdna(
            @ApiIgnore Authentication authentication,
            @RequestBody @Valid cz.mendelu.dnaAnalyser.analyse.zdna.ZdnaController.ZdnaRequest request) {

        Zdna zdna = zdnaService.executeZdnaAnalysis(
                Zdna.builder()
                        .sequence(sequenceService.findOne(request.getSequence()))
                        .owner(JwtAuthenticationUtils.toUser(authentication))
                        .tags(request.getTags())
                        .minSequenceSize(request.getMinSequenceSize())
                        .threshold(request.getThreshold())
                        .score_gc(request.getScore_gc())
                        .score_gtac(request.getScore_gtac())
                        .score_at(request.getScore_at())
                        .score_oth(request.getScore_oth())
                        .build()
        );

        return Response.response(zdna);
    }

    @GetMapping
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get page with Z-DNA analysis.")
    public ResponsePage<Zdna> getPageWithZdnaAnalysis(
            PaginationParam paginationParam,
            SortParam sortParam,
            TagFilterParam tagFilterParam
    ) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<Zdna> zdnaAnalysePage = zdnaService.findAll(pageable, tagFilterParam.getTags());
        ResponsePage<Zdna> responsePage = ResponsePage.responsePage(zdnaAnalysePage);
        zdnaAnalysePage
                .forEach(zdna -> responsePage.setDownloadToken(jwtTokenService.generateDownloadToken()));
        return responsePage;
    }


    @GetMapping("/{id}/zdnas")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get one Z-DNA analysis by ID.")
    public ResponsePage<ZdnaAnalyseResult> getZdnaList(
            @PathVariable UUID id,
            Integer sequenceStart,
            Integer sequenceLength,
            PaginationParam paginationParam,
            SortParam sortParam)
    {

        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<ZdnaAnalyseResult> zdnaPage = zdnaService.returnZdnaAnalysis(id, sequenceStart, sequenceLength, pageable);
        return ResponsePage.responsePage(zdnaPage);
    }

    @GetMapping("/{id}/analysis")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get one Z-DNA analysis by ID.")
    public Response<Zdna> getZdnaAnalysis(
            @PathVariable UUID id) {
        Zdna zdna = zdnaService.findOne(id);
        Response<Zdna> response = Response.response(zdna);
        response.setDownloadToken(jwtTokenService.generateDownloadToken());

        return response;
    }

    @GetMapping("/{id}/heatmap")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get Z-DNA heatmap by Id.")
    public Heatmap getZdnaHeatmap(
            @PathVariable UUID id,
            Integer segments,
            Integer from,
            Integer to)
    {
        return zdnaService.getHeatmap(id, segments, from, to);
    }

    @GetMapping("/tag")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get all tags defined for sequences")
    public ResponsePage<String> getPageWithTags(
            PaginationParam paginationParam,
            SortParam sortParam
    ) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<String> sequencePage = zdnaService.findAllTags(pageable);
        return ResponsePage.responsePage(sequencePage);
    }

    @GetMapping("/{id}/average/length")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get average Z-DNA length for single analysis")
    public Integer getAvgZdnaLen(@PathVariable UUID id)
    {
        return zdnaService.avgZdnaLen(id);
    }

    @GetMapping("/{id}/total/length")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get total Z-DNA length for single analysis")
    public Integer getTotalZdnaLen(@PathVariable UUID id)
    {
        return zdnaService.totalZdnaLen(id);
    }

    @PutMapping("/{id}/tags")
    @ResponseStatus(HttpStatus.ACCEPTED)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Modify tags")
    public Response<Zdna> modifyTags(
            @PathVariable UUID id,
            @RequestBody TagModifyRequest request) {
        Zdna analyse = zdnaService.modifyTags(id, request.getTags());
        return Response.response(analyse);
    }

    @DeleteMapping("/{id}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Delete one Z-DNA analysis by ID")
    public void deleteZdnaRAnalysis(@PathVariable UUID id) {

        zdnaService.delete(id);
    }

    @GetMapping(
            path = "/{id}/zdna.csv",
            produces = MediaType.TEXT_PLAIN_VALUE
    )
    @ResponseStatus(HttpStatus.OK)
    @ApiOperation("Get CSV file.")
    public ResponseEntity<String> getCSV(@PathVariable UUID id)
    {
        String csv = exporterService.exportData(id, Analysis.ZDNA, Output.CSV);

        HttpHeaders headers = new HttpHeaders();
        headers.setContentDisposition(ContentDisposition
                .builder("attachment")
                .filename(id + ".csv")
                .build());
        headers.setContentType(MediaType.valueOf("text/csv"));
        headers.setContentLength(csv.length());

        return new ResponseEntity<>(csv, headers, HttpStatus.OK);
    }

    @GetMapping(
            path = "/{id}/zdna.bedgraph",
            produces = MediaType.TEXT_PLAIN_VALUE
    )
    @ResponseStatus(HttpStatus.OK)
    @ApiOperation("Get BEDGRAPH file.")
    public ResponseEntity<String> getBedgraph(@PathVariable UUID id)
    {
        String bedgraph = exporterService.exportData(id, Analysis.ZDNA, Output.BEDGRAPH);

        HttpHeaders headers = new HttpHeaders();
        headers.setContentDisposition(ContentDisposition
                .builder("attachment")
                .filename(id + ".bedgraph")
                .build());
        headers.setContentType(MediaType.valueOf("text/bedgraph"));
        headers.setContentLength(bedgraph.length());

        return new ResponseEntity<>(bedgraph, headers, HttpStatus.OK);
    }
}

